/*  ARRAY3.C
 *  demonstrates use of memory log with 3-dimensional
 *  array allocation and free functions
 *  (also demonstrates floating point frailties)
 *  last mod.: 11-FEB-91
 */

#include <STDIO.H>
#include <STDLIB.H>
#include <STRING.H>

#include "L_MEM.H"

#define N1 20
#define N2 30
#define N3 20

int err_flag;
char msg[100];

typedef double huge * far * far *Array_3_double;

void main(int argc,
          char **argv)
{
Array_3_double a1, a2, a3;              /*  see typedef above  */
double val, init_val = atof(argv[1]);   /*  a3 initialization value  */
int i, j, k, m;
int print = ( argv[2][0] == 'P' || argv[2][0] == 'p' );
                                        /*  print log and heap if TRUE  */
FILE *output;                           /*  log output  */

if ( argc == 1 )
    {
    printf("Use: ARRAY3 initialization_value [P [log_file]]."
         "\n\"initialization_value\" is any floating point value."
         "\nUse \"P\" to print the heap and the memory allocation log"
         "\n(optionally to disk file \"log_file\").\n");
    exit(0);
    }

if ( print )
    {
    if ( argc < 4 )
        output = stdprn;
    else if ( ( output = fopen(argv[3],"wt") ) == NULL )
        {
        printf("\nUnable to open file %s",argv[3]);
        exit(1);
        }
    }

mem_log_on();
/*  array_expansion_off();
 *  activatation of this statement suppresses recording in the log
 *  of memory block operations constituting array operations
 */
strcpy(msg,"Program is ");
strcat(msg,argv[0]);
mem_log_msg(msg,&err_flag);

sprintf(msg,"Initial value for a3 is %f.",init_val);
mem_log_msg(msg,&err_flag);

/*  set termination on detection of a critical error  */
crit_err_termination_on();
/* if a critical error is detected send report 1 to printer  */
set_crit_err_termination_output(PRINTER_OUTPUT,REPORT_1_ONLY);

/*  begin with something ridiculous  */
array3_free(NULL,&err_flag);

/*  now allocate three arrays  */
printf("\nNow allocating three 3-dimensional arrays of doubles, "
       "a1, a2 and a3.");

/*  note type casting of return value to Array_3_double,
 *  which is typedef'd above as: double huge * far * far *
 *  i.e. a far pointer to a far pointer to a huge pointer to a double
 */
a1 = (Array_3_double)array3_alloc(N1,N2,N3,
    sizeof(double),NULL,&err_flag);
if ( err_flag )
    {
    printf("\n%s",mem_err_msg(err_flag,msg));
    printf("\nUnable to allocate a1.\n");
    exit(2);
    }

/*  NULL means: don't initialize the array  */
a2 = (Array_3_double)array3_alloc(N1,N2,N3,
    sizeof(double),NULL,&err_flag);
if ( err_flag )
    {
    printf("\n%s",mem_err_msg(err_flag,msg));
    printf("\nUnable to allocate a2.\n");
    exit(3);
    }

/*  note initialization to given value, namely, init_val  */
a3 = (Array_3_double)array3_alloc(N1,N2,N3,sizeof(double),
    (char *)&init_val,&err_flag);
if ( err_flag )
    {
    printf("\n%s",mem_err_msg(err_flag,msg));
    printf("\nUnable to allocate a3.\n");
    exit(4);
    }

/*  initialze arrays a2 and a3 subsequent to allocation  */
printf("\nNow initializing %dx%dx%d arrays a1 and a2 ",N1,N2,N3);
for ( i=0; i<N1; i++ )
    {
    printf("-");
    for ( j=0; j<N2; j++)
        {
        for ( k=0; k<N3; k++ )
            {
            a1[i][j][k] = (double)i + j;
            a2[i][j][k] = i - k - init_val;
            /*  check previous initialization of a3  */
            if ( a3[i][j][k] != init_val )
                {
                printf("\na3[%d][%d][%d] = %f, should be %f.",
                    i,j,k,a3[i][j][k],init_val);
                exit(5);
                }
            }
        }
    }

printf("\nNow checking results of calculation ");
for ( i=0; i<N1; i++ )
    {
    printf("-");
    for ( j=0; j<N2; j++)
        {
        for ( k=0; k<N3; k++ )
            {
            val = a1[i][j][k] - a2[i][j][k] - j - k;
            if ( a3[i][j][k] != val )
                {
                /*  inequality may occur due to slight inaccuracy
                 *  in representation of floating point values
                 */
                printf("\nUnexpected result:");
                printf("\na3[%d][%d][%d] = %f",i,j,k,a3[i][j][k]);
                printf("\nval = a1[%d][%d][%d] - a2[%d][%d][%d]"
                    " - %d - %d = %f",i,j,k,i,j,k,j,k,
                    a1[i][j][k]-a2[i][j][k]-j-k);
                printf("\nval - a3[i][j][k] = %f, and val %s= a3[i][j][k]",
                    val-a3[i][j][k],(val==a3[i][j][k]?"":"!"));
                /*  display the actual hexadecimal representation of
                 *  the two floating point values
                 */
                printf("\na3[%d][%d][%d]: ",i,j,k);
                for ( m=0; m<sizeof(double); m++ )
                    printf("%02X ",0xFF&*((char far *)&a3[i][j][k]+m));
                printf("\nval:         ");
                for ( m=0; m<sizeof(double); m++ )
                    printf("%02X ",0xFF&*((char far *)&val+m));
                exit(6);
                }
            }
        }
    }

printf("\nCalculation and comparisons check out OK.\n");

/*  check available memory (write in log)  */
mem_max(&err_flag);

/*  do more ridiculous things  */

/*  attempt to allocate an impossibly large array  */
array3_alloc(100,100,100,
    sizeof(double),NULL,&err_flag);

/*  attempt to allocate an array of elements of size 0  */
array3_alloc(100,100,100,0,NULL,&err_flag);

/*  allocate a memory block  */
mem_alloc(1000L,&err_flag);

if ( print )
    {
    /*  dump the heap  */
    if ( output == stdprn )
        printf("\nNow dumping far heap ...");
    else
        printf("\nNow dumping far heap to %s.",strupr(argv[3]));
    far_heap_dump(NULL,-1,32L,output,TRUE,&err_flag);
    /*  TRUE = formfeed  */
    if ( err_flag )
        {
        printf("\n%s",mem_err_msg(err_flag,msg));
        exit(7);
        }
    }

/*  free the first and second arrays specifically  */
array3_free(a1,&err_flag);
array3_free(a2,&err_flag);

/*  attempt to free a1 and a2 a second time  */
array3_free(a1,&err_flag);
array3_free(a2,&err_flag);

/*  now free allocated memory block and third array together  */
mem_free_all(0,&err_flag);

if ( print )
    {
    if ( output == stdprn )
        printf("\nNow printing reports ...");
    else
        printf("\nNow writing reports to %s.",strupr(argv[3]));
    mem_log_report_1(output,1,TRUE,&err_flag);
    if ( !err_flag )
        mem_log_report_2(output,0,TRUE,&err_flag);
    if ( err_flag )
        {
        printf("\n%s",mem_err_msg(err_flag,msg));
        exit(8);
        }
    printf("\n");
    }
else
    printf("\nThere are %u entries in the memory allocation log."
        "\nUse \"ARRAY3 init_val P\" to print the log or "
        "\n\"ARRAY3 init_val P log_file\" to write the log to disk.\n",
        mem_log_entries());
}
